from ..adapters import IAdapter
from .iprotocol import IProtocol

class Raw(IProtocol):
    def __init__(self, adapter: IAdapter) -> None:
        """
        Raw device, no presentation and application layers

        Parameters
        ----------
        adapter : IAdapter
        """
        super().__init__(adapter)

    def write(self, data : bytearray):
        self._adapter.write(data)

    def query(self, data : bytearray) -> bytearray:
        self._adapter.flushRead()
        self.write(data)
        return self.read()

    def read(self) -> bytearray:
        return self._adapter.read()


# adapters.py
# Sébastien Deriaz
# 06.05.2023
#
# Adapters provide a common abstraction for the media layers (physical + data link + network)
# The following classes are provided, which all are derived from the main Adapter class 
#   - IP
#   - Serial
#   - VISA
# 
# Note that technically VISA is not part of the media layer, only USB is.
# This is a limitation as it is to this day not possible to communication "raw"
# with a device through USB yet
#
#
#
#
# Timeout system
# We can differentiate two kinds of timeout :
#   - transmission timeout (aka "timeout"): the time it takes for a device to start transmitting what we expect
#   - continuation timeout : the time it takes for a device to continue sending the requested data

from abc import abstractmethod, ABC

class IAdapter(ABC):
    @abstractmethod
    def __init__(self, descriptor, *args):
        pass

    @abstractmethod
    def flushRead(self):
        """
        Flush the input buffer
        """
        pass

    @abstractmethod
    def open(self):
        """
        Start communication with the device
        """
        pass

    @abstractmethod
    def close(self):
        """
        Stop communication with the device
        """
        pass
            
    @abstractmethod
    def write(self, data : bytearray):
        """
        Send data to the device
        """
        pass
    
    @abstractmethod
    def read(self, timeout=None, continuation_timeout=None, until_char=None) -> bytearray:
        """
        Read data from the device
        """
        pass

    @abstractmethod
    def query(self, data : bytearray, timeout=None, continuation_timeout=None) -> bytearray:
        """
        Shortcut function that combines
        - flush_read
        - write
        - read
        """
        pass



from pyvisa import ResourceManager

from .iadapter import IAdapter

class VISA(IAdapter):
    def __init__(self, descriptor : str):
        """
        USB VISA stack adapter

        Parameters
        ----------
        descriptor : str
            IP description
        """
        self._rm = ResourceManager()
        self._inst = self._rm.open_resource(descriptor)
        self._inst.write_termination = ''
        self._inst.read_termination = ''

    def list_devices(self=None):
        """
        Returns a list of VISA devices
        """
        rm = ResourceManager()

        return rm.list_resources()

    def flushRead(self):
        pass

    def open(self):
        self._inst.open()

    def close(self):
        self._inst.close()
            
    def write(self, data : bytearray):
        self._inst.write(data.decode('ASCII'))
    
    def read(self):
        return self._inst.read()

    def query(self, data : bytearray, timeout=None, continuation_timeout=None) -> bytearray:
        """
        Shortcut function that combines
        - flush_read
        - write
        - read
        """
        # TODO : implement timeouts
        return self._inst.query(data.encode('ASCII'))


from ..adapters import IAdapter, IP
from .iprotocol import IProtocol


class SCPI(IProtocol):
    DEFAULT_PORT = 5025
    def __init__(self, adapter: IAdapter, end = '\n') -> None:
        """
        SDP (Syndesi Device Protocol) compatible device

        Parameters
        ----------
        wrapper : Wrapper
        """
        super().__init__(adapter)
        self._end = end

        if isinstance(self._adapter, IP):
            self._adapter.set_default_port(self.DEFAULT_PORT)

    def _to_bytearray(self, command):
        if isinstance(command, str):
            return command.encode('ASCII')
        elif isinstance(command, bytes) or isinstance(command, bytearray):
            return command
        else:
            raise ValueError(f'Invalid command type : {type(command)}')

    def _formatCommand(self, command):
        return command + self._end

    def _unformatCommand(self, payload):
        return payload.replace(self._end, '')
    
    def _checkCommand(self, command : bytearray):
        for c in ['\n', '\r']:
            if c in command:
                raise ValueError(f"Invalid char '{c}' in command")

    def write(self, command : bytearray) -> None:
        self._checkCommand(command)
        command = self._to_bytearray(command)
        self._adapter.write(self._formatCommand(command))

    def query(self, command : str) -> str:
        data = self._to_bytearray(command)
        self._adapter.flushRead()
        self.write(data)
        return self.read()

    def read(self) -> str:
        data = self._adapter.read().decode('ASCII')
        return self._unformatCommand(data)


# Syndesi ID
#
# Sébastien Deriaz
# 14.11.2022
from enum import Enum

class AddressType(Enum):
        NONE = 0
        IPV4 = 1
        IPV6 = 2

class SyndesiID:
    def __init__(self, addressType : AddressType):
        """
        Syndesi ID instance

        Parameters
        ----------

        """
        self.addressType = addressType


from ..adapters import IAdapter
from .iprotocol import IProtocol


# This is a temporary class to manage continuously streaming devices (Arduino for datalogging, some RS-232 multimeters, etc...)

class RawStream(IProtocol):
    def __init__(self, adapter: IAdapter, delimiter = b'\n') -> None:
        """
        Continuously streaming device with specified delimiter

        Parameters
        ----------
        adapter : IAdapter
        delimiter : bytearray
            Sample delimiter
        """
        super().__init__(adapter)

        self._buffer = b''
        self._delimiter = delimiter

    def flushRead(self):
        self._adapter.flushRead()

    def write(self, data : bytearray):
        self._adapter.write(data)

    def query(self, data : bytearray) -> bytearray:
        self._adapter.flushRead()
        self.write(data)
        return self.read()

    def read(self, include_last=False):
        """
        Return a list of samples

        Parameters
        ----------
        include_last : bool
            Include the last sample, which is probably incomplete, False by default
        """
        self._buffer += self._adapter.read()
        samples = self._buffer.split(self._delimiter)
        if include_last:
            self._buffer = b''
            return samples
        else:
            # Put the last one back in the buffer as it probably isn't complete yet
            self._buffer = samples[-1]
            if len(samples) > 1:
                # If there's more than one sample, output all except the last one
                return samples[:-1]
            else:
                # Otherwise output nothing
                return []



import socket
from enum import Enum
from .iadapter import IAdapter

class IP(IAdapter):
    class Status(Enum):
        DISCONNECTED = 0
        CONNECTED = 1
    class Protocol(Enum):
        TCP = 0
        UDP = 1
    def __init__(self, descriptor : str, port = None, transport : Protocol = Protocol.TCP):
        """
        IP stack adapter

        Parameters
        ----------
        descriptor : str
            IP description
        port : int
            port used (optional, can be specified in descriptor)
        transport : Transport
            Transport protocol, TCP or UDP
        """
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(5)
        self._ip = descriptor # TODO : update this
        self._port = port
        self._status = self.Status.DISCONNECTED

    def set_default_port(self, port):
        """
        Sets IP port if no port has been set yet.

        This way, the user can leave the port empty
        and the driver/protocol can specify it later
        
        Parameters
        ----------
        port : int
        """
        if self._port is None:
            self._port = port


    def flushRead(self):
        return super().flushRead()

    def open(self):
        self._socket.connect((self._ip, self._port))
        self._status = self.Status.CONNECTED

    def close(self):
        self._socket.close()
            
    def write(self, data : bytearray):
        if self._status == self.Status.DISCONNECTED:
            self.open()
        self._socket.send(data)

    def read(self):
        if self._status == self.Status.DISCONNECTED:
            self.open()
        
        self._socket.settimeout(10)

        buffer = b''
        while True:
            try:
                recv = self._socket.recv(10)
                self._socket.settimeout(0.05)
                buffer += recv
            except socket.timeout as e:
                break
        return buffer

    def query(self, data : bytearray):
        self.flushRead()
        self.write(data)
        return self.read()
        



# Syndsi Device class
#
# Sébastien Deriaz
# 14.11.2022
from .payload import Payload
from .network import network
from .frame import OutboundFrame, InboundFrame
from .sdid import SyndesiID, AddressType

class Device:
    def __init__(self, descriptor):
        """
        New device instance
        """


    def send(self, payload : Payload):
        
        ID = SyndesiID(AddressType.IPV4)

        frame = OutboundFrame(payload, ID)

        outputFrame = network.request(frame)

        return outputFrame.payload


from .iprotocol import IProtocol
from ..adapters import IAdapter


class RawCommands(IProtocol):
    def __init__(self, adapter : IAdapter, end=b'\n', format_response=True) -> None:
        """
        Command-based protocol, with LF, CR or CRLF termination

        No presentation or application layers

        Parameters
        ----------
        adapter : IAdapter
        end : bytearray
            Command termination, '\n' by default
        format_response : bool
            Apply formatting to the response (i.e removing the termination)
        """
        super().__init__(adapter)

        if not isinstance(end, bytes):
            raise ValueError(f"end argument must be of type bytes, not {type(end)}")
        self._end = end
        self._response_formatting = format_response

    def _to_bytearray(self, command) -> bytearray:
        if isinstance(command, str):
            return command.encode('ASCII')
        elif isinstance(command, bytes) or isinstance(command, bytearray):
            return command
        else:
            raise ValueError(f'Invalid command type : {type(command)}')

    def _format_command(self, command : bytearray) -> bytearray:
        return command + self._end
    
    def _format_response(self, response : bytearray) -> bytearray:
        if response.endswith(self._end):
            response = response[:-len(self._end)]
        return response

    def write(self, command : bytearray):
        command = self._to_bytearray(command)
        self._adapter.write(self._format_command(command))

    def query(self, command : bytearray) -> bytearray:
        data = self._to_bytearray(command)
        self._adapter.flushRead()
        self.write(data)
        return self.read()

    def read(self) -> bytearray:
        if self._response_formatting:
            return self._format_response(self._adapter.read())
        else:
            return self._adapter.read()


# Syndesi Frame
#
# Sébastien Deriaz
# 14.11.2022

from enum import Enum
from .sdid import SyndesiID, AddressType
from .payload import Payload
from struct import pack, unpack, calcsize

def _bit_to_bool(buffer, bit):
        return True if buffer & (1 << bit) else False

class NetworkHeader:
    length = 1
    def __init__(self, routing=True, follow=False, error=True):
        """
        Network header
        """
        self.routing = routing
        self.follow = follow
        self.error = error

    def _check_values(self):
        """
        Check values for correct types and values
        """
        assert isinstance(self.routing, bool), f'"routing" has the wrong type {type(self.routing)} instead of bool'
        assert isinstance(self.follow, bool), f'"follow" has the wrong type {type(self.follow)} instead of bool'
        assert isinstance(self.error, bool), f'"error" has the wrong type {type(self.error)} instead of bool'

    def value(self):
        """
        Export network header as byte array
        """
        self._check_values()

        value = self.routing | (self.follow << 1) | (self.error << 2)

        return value
    
    def parse(self, buffer):
        """
        Parse the header from a buffer

        Parameters
        ----------
        buffer : int
        """

        self.routing = _bit_to_bool(buffer, 0)
        self.follow = _bit_to_bool(buffer, 1)
        self.error = _bit_to_bool(buffer, 2)

class ErrorCode(Enum):
    NO_ERROR = 0,
    NO_INTERPETER = 1,
    INVALID_PAYLOAD = 2



class OutboundFrame():
    def __init__(self, payload : Payload, ID : SyndesiID):
        """
        Outbound Frame instance (host -> device)

        Parameters
        ----------
        payload : Payload
            The payload to send
        ID : SyndesiID
            Device's ID
            
        """
        self.ID = ID
        self._payload = payload

    def data(self):
        """
        Return frame raw data
        """
        STRUCT_FORMAT = lambda pl : f'!BH{pl}s'
        # Create network header
        header = NetworkHeader()
        header.follow = False
        header.routing = False
        header.error = False
        # Payload
        payload_buffer = self._payload.value()
        # Frame length
        length = len(payload_buffer) # Length of payload

        buffer = pack(STRUCT_FORMAT(length), header.value(), length, payload_buffer)

        return buffer

class InboundFrame():
    def __init__(self, buffer : bytearray, requestFrame : OutboundFrame) -> None:
        """
        Inbound frame instance (device -> host)
        
        Parameters
        ----------
        buffer : bytearray
            The buffer to parse
        ID : SyndesiID
            Device's ID
        requestFrame : OutboundFrame
            The frame sent for the request, this is to identify the confirm payload
        """
        self.ID = requestFrame.ID
        self.isErrorFrame = False

        # Parsing

        # Read the header
        HEADER_STRUCT_FORMAT = '!BH'
        header_size = calcsize(HEADER_STRUCT_FORMAT)
        
        header_byte, length_or_error = unpack(HEADER_STRUCT_FORMAT, buffer[:header_size])

        # Header
        self.header = NetworkHeader()
        self.header.parse(header_byte)

        self.isErrorFrame = True if self.header.error else False
        if self.isErrorFrame:
            # Error frame, the rest is the error code
            self.errorCode = ErrorCode(length_or_error)
        else:
            # Read the payload
            payloadBuffer = buffer[header_size:header_size+length_or_error]
            self.payload = requestFrame._payload._reply(payloadBuffer)


# Syndesi network
#
# Sébastien Deriaz
# 14.11.2022

import socket
from .frame import OutboundFrame, InboundFrame

class Network:
    def __init__(self):
        """
        Syndesi Network instance
        """


    def request(self, requestFrame : OutboundFrame):
        """
        Send the provided frame

        Parameters
        ----------
        frame : Frame
        """

        PSDU = requestFrame.data()
        #print(f"Sending frame {PSDU}")

        # TODO : Add protocol check (only IP for now)
        # TODO : Change IP address
        HOST = "192.168.1.67"
        PORT = 2608

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((HOST, PORT))
            s.sendall(PSDU)
            confirm_data = s.recv(1024)
            return InboundFrame(confirm_data, requestFrame)
        
        return None

network = Network()


# Sébastien Deriaz
# 21.02.2023
# 
# VISA descriptor
# NOTE : Not sure this is the right approach, check this later

from .descriptor import Descriptor

import pyvisa

rm = pyvisa.ResourceManager()

class VISA(Descriptor):
    def __init__(self, descriptor : str) -> None:
        super().__init__()

        descriptors = rm.list_resources()

        if descriptor not in descriptors:
            raise RuntimeError(f'Descriptor "{descriptor} unavailable". Availables are : {descriptors}')

        self._inst = rm.open_resource(descriptor)

    def read(self):
        return self._inst.read()

    def write(self, data):
        self._inst.write(data)

    def __del__(self):
        self._inst.close()


from setuptools import setup, find_packages

VERSION = '0.1.1'
DESCRIPTION = 'Syndesi'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Setting up
setup(
    name="syndesi",
    version=VERSION,
    author="Sebastien Deriaz",
    author_email="sebastien.deriaz1@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    scripts=['bin/syndesi'],
    packages=find_packages(),
    install_requires=[''],
    keywords=['python', 'syndesi', 'interface', 'ethernet'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)



from .iadapter import IAdapter

import serial

class Serial(IAdapter):
    def __init__(self, port : str, baudrate=115200):
        """
        Serial communication adapter

        Parameters
        ----------
        port : str
            Serial port (COMx or ttyACMx)
        """
        self._port = serial.Serial(port=port, baudrate=baudrate)

    def flushRead(self):
        self._port.flush()

    def open(self):
        self._port.open()

    def close(self):
        self._port.close()
            
    def write(self, data : bytearray):
        self._port.write(data)
    
    def read(self):
        # TODO : Implement timeout strategy
        return self._port.read_all()

    def query(self, data : bytearray):
        self.flushRead()
        self.write(data)
        return self.read()


from ..adapters import IAdapter
from .iprotocol import IProtocol


class SDP(IProtocol):
    def __init__(self, adapter: IAdapter) -> None:
        """
        SDP (Syndesi Device Protocol) compatible device

        Parameters
        ----------
        wrapper : Wrapper
        """
        super().__init__(adapter)